module Kleben where


import Monad
import Winkel
import Correct

---------------------------------------------------------------------------

voll :: Ecke a
voll =  let w = 1 
	in Ecke { sichtbar = w
		, alle = [w]
		, info = [] 
		}

versteckter :: Winkels -> Ecke a
versteckter ws = Ecke { sichtbar = 0
		      , alle = ws
		      -- ein bichen gefhrlich, aber das Feld wird
		      -- erst nachtrglich gesetzt, siehe unten
		      , info = undefined
		      }

---------------------------------------------------------------------------

zerlegung :: MonadPlus m => [a] -> m (a, [a], a)
-- zerlegung [1..5] = (1, [2..4], 5)
-- das sollte man eigentlich mit double ended queues
-- und nicht mit listen implementieren
zerlegung (x : xs) = do
    (y : ys) <- return $ reverse xs 
    return (x, reverse ys, y)
zerlegung [] = mzero

-- zwei Streckenzge verkleben entlang jeweils
-- der seite zwischen dem letzten und ersten Winkel
klebe :: (Correct i, MonadPlus m)
      => Streckenzug i -> Streckenzug i -> m (Streckenzug i)
klebe (Streckenzug p) (Streckenzug q) = do
    (p_erster, p_mitte, p_letzter) <- zerlegung p
    (q_erster, q_mitte, q_letzter) <- zerlegung q

    let qp = ( q_letzter * p_erster ) -- Reihenfolge ist wichtig
    guard $ correct (info qp)

    let pq = ( p_letzter * q_erster )
	     { alle = alle p_letzter ++ [1] ++ alle q_erster }
		   -- das ist ein "update using field labels", siehe
		   -- http://haskell.org/onlinereport/exps.html#sect3.15.3
    guard $ correct (info pq)

    let k =  [ qp ] ++ p_mitte ++ [ pq ] ++ q_mitte
    ohne_spitzen $ Streckenzug $ k

-- m ist eine Variable fr einen Typkonstruktor,
-- da setzen wir spter Maybe oder [] ein.
ohne_spitzen :: (Correct i, MonadPlus m) 
	     => Streckenzug i -> m (Streckenzug i)
ohne_spitzen (Streckenzug p) =
    case span ( < voll ) p of
	(gut, []) -> return $ Streckenzug gut

        (gut, a : rest) | a > voll -- mehr als Vollwinkel darf nicht sein
	    -> mzero		   -- wird Nothing oder leere Liste

	([], a : rest) | a == voll -- Spitze am Anfang
	    -> do 
		  guard $ correct (info a) -- brauchen wir das?

		  (erster, mitte, letzter) <- zerlegung rest
		  let umweg =  alle a
		      neu = ( letzter + versteckter umweg + erster )
			    { info = info letzter ++ info erster }
		  ohne_spitzen $ Streckenzug $ neu : mitte

	(gut, a : rest) -- Spitze mittendrin, rotieren
	    -> ohne_spitzen $ Streckenzug $ a : rest ++ gut

---------------------------------------------------------------------------

