module Morph where

import Array

-- in /usr/local/share/hugs/lib/Array.hs kann man sich anschauen,
-- was mit Arrays alles geht (und zwar ohne Tricks nicht viel)

type Morph = Array Char String

-- so baut man ein Array:

-- Fibonacci
f = array ('0', '1')		      -- Indexbereich
	  [ ('0', "01"), ('1', "0") ] -- Liste von (Index, Wert) 

-- Thue
t = array ('0', '1') [ ('0', "01"), ('1', "10") ] 

-- Morse
m = array ('0', '2') [ ('0', "012"), ('1', "02"), ('2', "1") ] 

------------------------------------------------------------------

morph :: Morph -> String -> String
morph m [] = []
morph m (x : xs) = m ! x ++ morph m xs

d0l :: Morph -> String
-- implizite Annahme: wir beginnen mit 0
-- und  m (0)  mu 0 als Prfix haben
d0l m = '0' : w 
    where '0' : rest = m ! '0'
	  w = rest ++ morph m w  -- das wird eine unendliche Liste!

fib, morse, thue :: String
fib = d0l f
thue = d0l t
morse = d0l m

-- Aufgaben: Prfe durch ein Programm, 
-- da  thue  Kubus-frei ist, und  morse  quadratfrei.

-- Allgemein: finde ein Entscheidungsverfahren fr:
--   Eingabe: ein (H)D0L-System S und ein Muster m
--   Frage  : kommt m im von S beschriebenen Wort vor?

-- und noch besser: ein Verfahren fr:
--   Eingabe: ein Muster m
--   Ausgabe: ein (H)D0L-System, das m vermeidet,
--            oder ein Beweis dafr, da so etwas nicht existiert

--------------------------------------------------------------------------

hd0l :: (Morph, Morph) -> String
hd0l (f, h) = morph h (d0l f)

h :: Morph
h = array ('0', '2') [ ('0', "011"), ('1', "01"), ('2', "0") ] 

-- Beweise  hd0l (m, h) == d0l t

-- Allgemein: finde ein Entscheidungsverfahren fr:
--   Eingabe: zwei HD0L-Systeme
--   Frage  : erzeugen beide das gleiche unendliche Wort?

-- Bemerkung: fr D0L-Systeme ist das entscheidbar, aber nicht trivial

--------------------------------------------------------------------------

main = do
    let l = 65
    putStrLn $ "Fibonacci  : " ++ take l ( d0l f )
    putStrLn $ "Morse      : " ++ take l ( d0l m )
    putStrLn $ "Thue       : " ++ take l ( d0l t )
    putStrLn $ "hd0l (m,h) : " ++ take l ( hd0l (m, h) )

