module PCP where

import Rose

--import Form
--import Bild
--import Ding

import BFS
import Sort
import List (tails)

type PCP a = [ ([a],[a]) ]

ex :: PCP Char
-- lsungslnge 205 (Lorentz)
ex = [ ("000", "0"), ("0", "111"), ("11", "0"), ("10", "100") ]

lr =  map ( \ (u, v) -> (reverse u, reverse v) ) 
ou =  map ( \ (u, v) -> (v, u) )

xe = lr ex

-- lsungslngen 160 (Lorentz)
kn = [ ("100","1"), ("00","110"), ("1","00"), ("1", "011") ]
un = [ ("01","0"), ("00","011"), ("1","101"), ("001", "1") ]
nu = ou un

-- lnge 119 (Lorentz, Waldmann)
lw = ou [ ("1","0"), ("10","1011"), ("1101","1") ]

f u v = [("0", "0" ++ v), ("1","0"), (u ++ "1", "1")]


fib8 = f "1010010" "1"

fib7 = [("0","01"), ("1","0"), ("0010010","1")] -- no
fib6 = [("0","01"), ("1","0"), ("001001","1")]
fib5 = [("0","01"), ("1","0"), ("00100","1")] -- no
fib4 = [("0","01"), ("1","0"), ("1010","1")]  -- no
fib3 = [("0","01"), ("1","0"), ("101","1")]


data State a = State { sol  :: [ Int ]
		     -- (oben, unten), einer von beiden ist leer
		     , diff :: ( [a], [a] ) 
		     -- schranke fr krzeste erreichbare Lsung
		     , bound :: Int 
		     }

ldiff :: State a -> Int
ldiff s = let (o, u) = diff s in length o - length u

instance Show a => Show (State a) where
    showsPrec p s = showString (
		     "\n{ sol = " ++ show (sol s)
		  ++ "\n, length sol = " ++ show (length (sol s))
		  ++ "\n, diff = " ++ show (diff s)
		  ++ "\n, length diff = " ++ show (ldiff s)
		  ++ "\n, bound = " ++ show ( bound s )
		  ++ "\n}"
	      )

{-
instance Form a => Form (State a) where
    form s = let (oben, unten) = diff s
	     in  ber ( nebens $ map form oben )
		      ( nebens $ map form unten )
-}

instance Eq a => Eq (State a) where
    s1 == s2  =   diff s1 == diff s2

instance Ord a => Ord (State a) where
    s1 `compare` s2  =   diff s1 `compare` diff s2

start :: State a
start = State { sol = [], diff = ([], []), bound = 0 }

-- alle Nachfolger eines Zustands
nach :: Eq a => PCP a -> State a -> [ State a ]
nach pcp s =
    [ State { sol = sol s ++ [k] 
	    , diff = (oben2, unten2)
	    , bound = bound s - 1 -- eventuell aber mehr
	    }
    | let (oben0, unten0) = diff s
    , (k, (u, v)) <- zip [1..] pcp
    , let (oben1, unten1) = (oben0 ++ u, unten0 ++ v)
    , let (oben2, unten2) = krze (oben1, unten1)
    , null oben2 || null unten2
    ]

-- entferne gemeinsamen Prfix
krze :: Eq a => ([a],[a]) -> ([a],[a]) 
krze (x : xs, y : ys) = 
    if x == y then krze (xs, ys)
	      else (x : xs, y : ys)
krze (xs, ys) = (xs, ys)

-- der komplette Suchbaum (unendlich!)

baum :: Eq a => PCP a -> Rose (State a)
baum pcp = generate start where
    generate s =
        Node { key = s
	     , children = map generate (nach pcp s)
	     }

-- eindeutige Ausdehnung:
straight :: Eq a => PCP a -> Int -> State a -> State a
straight pcp sch s =
    let (o, u) = diff s 
    in  if null o && null u then s else
    if  length (sol s) >=  sch then s else
    case nach pcp s of 
        [ t ] -> straight pcp sch t
	_     -> s

-- nur die Differenzen, die oben leer sind

-- obaum :: Eq a => PCP a -> Int -> Rose (State a)
obaum pcp sch = generate start where
    generate s0 =
        let s = s0 -- forcieren
	in  Node { key = s
		 , children = map generate	

    			  $ filter ( \ s -> bound s <= sch )
			  $ map ( schrank pcp )
			  $ filter ( \ s' -> 
			      let (oben, unten) = diff s' in null unten )

			  -- nur fr 119
			  -- $ filter ( \ x -> keine_zwei_nullen (fst (diff x)))

			  $ map (straight pcp sch)
			  $ nach pcp s
		 }

keine_zwei_nullen xs =
    and [ take 2 ys /= "00" | ys <- tails xs ]

extreme xs = (minimum xs, maximum xs)

schrank :: Eq a => PCP a -> State a -> State a
schrank pcp s = 
    let l = length (sol s)
	(o, u) = diff s; d = length o - length u
	(mi, ma) = extreme [ length o - length u | (o,u) <- pcp ]
	p = if d > 0
	    then d `div` (- mi)
	    else (-d) `div` ma
    in	s { bound = l + p }

-- schicht :: (Show a, Ord a) => PCP a -> (Int, [ State a ]) -> IO ()
schicht pcp (k, xs) = do
    putStrLn $ "schicht nummer " ++ show k
    putStrLn $ "anzahl der knoten " ++ show (length xs)

    let zs = filter ( (>0 ) . length . sol )
	   $ filter ( (== 0) . ldiff ) 
	   $ xs
    if not (null zs) 
       then do putStrLn $ "lsungen:"
	       sequence $ map print $ zs
	       error "foo"
       else return ()

    let ys = sortBy bound $ xs   
    putStrLn $ "Knoten mit kleinster Schranke " ++ show (head ys)
    putStrLn $ "Knoten mit grter Schranke " ++ show (last ys)

-- alles :: (Show a, Ord a) => PCP a -> Int -> IO ()
alles pcp sch = do
    putStrLn $ "suchbaum fr PCP " ++ show pcp
    putStrLn $ "fr Lsungen <= " ++ show sch
    sequence  $ map ( schicht pcp )
	      $ zip [0..]
	      $ bfs -- cut (take 200 . sortBy (bound . key))
	      $ obaum pcp sch
    return ()

